#!/usr/bin/perl
#
# Matthew J. Kidd (San Diego, CA)
#
# This software is released under the GNU General Public License GPLv3
# See: http://www.gnu.org/licenses/gpl.html for full license.
#
# Calculates the partnership payoff matrix by directly scanning ACBLscore game files.
# Also generates player and partnership statistics.
#
# Dependencies: JSON, Date-DayOfWeek, GD. JSON is shipped with most Perl installations.
# Run 'ppm install Date-DayOfWeek GD' to install Date-DayOfWeek and GD or just let the
# program install the packages when first needed.
#
# 25-Nov-2013 - Current release
# 05-Oct-2013 - Initial release
#
# See also ACBLmerge (http://lajollabridge.com/Software/ACBLmerge/ACBLmergeAbout.htm)

use strict;

my $VERSTR = '1.0.5';

my $DAYS = {'sun' => 0, 'mon' => 1, 'tue' => 2, 'wed' => 3, 'thu' => 4,
  'fri' => 5, 'sat' => 6};

# Board vulnerability (mod 16, first board is board 0, e.g. 16 or 32)
my @isVulNS = (0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1);
my @isVulEW = (1, 0, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 0, 0);

# Exit codes
my $EXIT_SUCCESS              =   0;
my $EXIT_ARG_PARSING          =  -1;
my $EXIT_NEED_PERL_PACKAGE    =  -2;
my $EXIT_NEED_MP_DATABASE     =  -3;
my $EXIT_BAD_MP_DATABASE      =  -4;
my $EXIT_MAKE_APPDIR_FAIL     = -12;

if ( scalar(@ARGV) == 0 ) {
  
  (my $bname = $0) =~ s/.*[\\\/]//;
	
  print "\n  Missing parameter(s)\n";
  
  print << "DONE";

  Usage: $bname [fname...] [-d dirname] [-pl plfname] [-pr prfname]
         [-pm pmfname] [-js jsfname] [-s list] [-dy list] [-y list]
         [-ul list] [-e list] [-mp list] [-mn #] [-ms #] [-lp #] [-ll]
         [-ip ipfname] [-io iofname] [-dp #] [-do #] [-px #] [-lc #]
         [-v #] [-vr]

  fname  : Input filename(s), i.e. ACBLscore game files
  dname  : Directory name (processes all games files in the directory)
           The -d option may be repeated to specify multiple directories

  Output files
    -pl : List of players and their statistics
    -pr : List of partnerships and their statistics
    -pm : Payoff matrix data (partnership-partnership interaction)
    -js : JavaScript variable assignment useful for HTML matrix display
    -ip : Payoff matrix image (regular presentation)
    -io : Payoff matrix image (over/underperform presentation) 
    
  General options
    -v  : Set verbosity level. 0 -> Quiet mode (minimize command line output)
          2 -> Detailed. Default is 1. Meaningful up to 4.
    -vr : Show version
    -sd : Dump full slam details in pair file 
    -ll : Dump a list of all upper limits encountered and how many events
          have each upper limit.
    -mp : Comma separated list of monthly masterpoint total columns to attach
          to player and partnerships file output. Format is YYMM, e.g. use
          1102,1310 to attach the Feb 2011 and Oct 2013 masterpoint holdings.
          If no list is supplied, masterpoint totals from start and end of
          time period covered by the game files will be attached.       
    
  Game file filtering options
    -s  : List of sessions (M = morning, A = afternoon, E = evening, L = late)
          to examine, e.g. AE. If not specified, all sessions are examined.
    -dy : Comma separated list of days to examine: mon, tue, wed, thu,
          fri, sat, sun. If not specified, all days are examined.
    -y  : Comma separated list of years to examine. Ranges may also be
          specified, e.g. 2007-2009,2011
          
  Event filtering options
    -ul : Comma separated list of masterpoint upper limits for events
          to examine. Use 'open' or 0 for events with no upper limit.       
    -e  : Comma separated list of event indices to examine. For example,
          if game files always have an open event and may optionally have
          a limited event as the second event then -e 1 will limit the
          examination to the open event. -ul is usually the better switch.
          
  Result filtering options
    -mn : Minimum number of sessions that a player must play in to be
          reported in the player list (-pl) output.
    -ms : Minimum number of sessions that a pair must play in to be reported
          in partnership list (-pr) and payoff matrix (-pm) output.
    -lp : Lowest player/partnership percentage to include in the player list
          (-pl) and partnership list (-pr) output.
          
  Image generation options
    -px : Number of pixels in each direction for each cell
    -dp : Dynamic range for regular payoff matrix color map 
    -do : Dynamic range for the over/underperform payoff matrix color map
          (defaults are both 30 => 20 to 80% and -30 to 30% respectively)
    -lc : Minimum boards in a pair-pair interaction to avoid red-X flagging

  Computes the partnership payoff matrix by directly scanning ACBLscore game
  files. Also computes player and partnership statistics. Can generate images
  of the payoff matrix and JavaScript support code for HTML presentation.

  Online documentation is located at:
  http://www.lajollabridge.com/Software/Payoff-Matrix/Payoff-Matrix-About.htm

DONE

  exit($EXIT_SUCCESS);
}

my (@fnames, %opt, $errmsg);

while ( my $arg = shift(@ARGV) ) {
  if ( substr($arg,0,1) ne '-' ) { push @fnames, $arg; next; };
  my $sw = substr($arg,1);
  if ($sw eq 'd') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-d switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }
    push @{$opt{'dirnames'}}, $arg;
  }
  elsif ($sw eq 'dp' || $sw eq 'do') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }    
    if ($arg !~ /^\d*\.?\d+|\d+$/ ) {
      print STDERR "\n-$sw switch requires a numeric argument.\n"; exit($EXIT_ARG_PARSING);
    }
    $opt{'dynRange1'} = $arg if $sw eq 'dp';
    $opt{'dynRange2'} = $arg if $sw eq 'do';
  }
  elsif ($sw eq 'dy') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-dy switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }
    requirePackage('Date::DayOfWeek', 'Selection by day of week (-dy) switch');
    $opt{'days'} = [0, 0, 0, 0, 0, 0, 0];
    foreach my $day (split /,/, $arg) {
      (my $lday = $day) =~ tr /A-Z/a-z/;
      if (!defined $DAYS->{$lday}) {
        print STDERR "\nBad day for -dy argument: $lday (use sun, mon, tue, wed, thu, fri, sat).\n";
        exit($EXIT_ARG_PARSING);
      }
      else {
        $opt{'days'}[$DAYS->{$lday}] = 1;
      }
    }
  }
  elsif ($sw eq 'e') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }
    my @ixlist = split(',', $arg);
    foreach my $ix (@ixlist) {
      if ($ix !~ /^\d+$/) {
        print STDERR "\nAll indices for -$sw switch must be an integers. Bad value: $ix\n";
        exit($EXIT_ARG_PARSING);
      }
      $opt{'eidx'}{$ix} = 1;
    }
  }
  elsif ($sw eq 'ul') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }
    my @ixlist = split(',', $arg);
    foreach my $ix (@ixlist) {
      if ($ix ne 'open' && $ix !~ /^\d+$/) {
        print STDERR "\nAll values for -$sw switch must be integers or 'open'. Bad value: $ix\n";
        exit($EXIT_ARG_PARSING);
      }
      $ix = 0 if $ix eq 'open';
      $opt{'ul'}{$ix} = undef;
    }
  }
  elsif ($sw eq 'pl' || $sw eq 'pr' || $sw eq 'pm' || $sw eq 'js' || $sw eq 'ip' || $sw eq 'io') {
    # Handle all the output filenames.
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }
    $opt{'plfname'} = $arg if $sw eq 'pl';
    $opt{'prfname'} = $arg if $sw eq 'pr';
    $opt{'pmfname'} = $arg if $sw eq 'pm';
    $opt{'jsfname'} = $arg if $sw eq 'js';
    $opt{'img1fname'} = $arg if $sw eq 'ip';
    $opt{'img2fname'} = $arg if $sw eq 'io';
  }
  elsif ($sw eq 'v') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }    
    if ($arg !~ /^\d+$/ ) {
      print STDERR "\n-$sw switch requires an integer argument.\n"; exit($EXIT_ARG_PARSING);
    }
    $opt{'verbosity'} = $arg;
  }
  elsif ($sw eq 'vr')  { $opt{'showver'} = 1; }
  elsif ($sw eq 'll')  { $opt{'dumpUpperLimits'} = 1; }
  elsif ($sw eq 'sd')  { $opt{'dumpSlamDetails'} = 1; }
  elsif ($sw eq 'ms') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }    
    if ($arg !~ /^\d+$/ ) {
      print STDERR "\n-$sw switch requires an integer argument.\n"; exit($EXIT_ARG_PARSING);
    }
    $opt{'minSessions'} = $arg;
  }
  elsif ($sw eq 'lp') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }    
    if ($arg !~ /^\-?\d+\.?(\d+)?$/ ) {
      print STDERR "\n-$sw switch requires an numeric argument.\n"; exit($EXIT_ARG_PARSING);
    }
    $opt{'minPercentage'} = $arg;
  }
  elsif ($sw eq 'mn') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }    
    if ($arg !~ /^\d+$/ ) {
      print STDERR "\n-$sw switch requires an integer argument.\n"; exit($EXIT_ARG_PARSING);
    }
    $opt{'minPlayerSessions'} = $arg;
  }
  elsif ($sw eq 'mp') {
    $opt{'mpcols'} = 1;
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') { unshift @ARGV, $arg; }
    else {
      foreach my $yymm (split /,/, $arg) {
        if ($yymm !~ /^\d{4}$/) {
          print STDERR "\n-$sw argument dates must have form YYMM. Bad value: $yymm\n";
          exit($EXIT_ARG_PARSING);
        }
        if (substr($yymm,-2) == 0 || substr($yymm,-2) > 12) {
          print STDERR "\n-$sw argument date month must be in range 1-12. Bad value: $yymm\n";
          exit($EXIT_ARG_PARSING);
        }
        $opt{'mpdates'}{$yymm} = undef;
      }
    }
  }
  elsif ($sw eq 'px') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }
    if ($arg !~ /^\d+$/ || $arg == 0) {
      print STDERR "\n-$sw switch argument must be a postive integer.\n"; exit($EXIT_ARG_PARSING);
    }
    $opt{'npixels'} = $arg;
  }
  elsif ($sw eq 'lc') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }
    if ($arg !~ /^\d+$/) {
      print STDERR "\n-$sw switch argument must be a postive integer.\n"; exit($EXIT_ARG_PARSING);
    }
    $opt{'lowconf'} = $arg;
  }
  elsif ($sw eq 's') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }
    if ($arg !~ /^[MAEL]+$/i) {
      print STDERR "\n-sw argument can only contain these letters: M, A, E, and L. Bad value: $arg\n";
      exit($EXIT_ARG_PARSING);
    }
    ($opt{'sessions'} = $arg) =~ tr/a-z/A-Z/;
  }
  elsif ($sw eq 'y') {
    $arg = shift(@ARGV);
    if (!defined $arg || substr($arg,0,1) eq '-') {
      print STDERR "\n-$sw switch is missing argument.\n"; exit($EXIT_ARG_PARSING);
    }
    foreach my $yearspan (split /,/, $arg) {
      if ( $yearspan =~ /^\d+$/ )  { $opt{'years'}{$yearspan} = undef; }
      elsif ( $yearspan =~ /^\d+-\d+$/ ) {
        my ($year1, $year2) = split /-/, $yearspan;
        for (my $year=$year1; $year <= $year2; $year++) {
          $opt{'years'}{$year} = undef;
        }
      }
      else {
        print STDERR "\nEach -$sw option list value must be an integer or an integer range.\n";
        print STDERR "$yearspan list value was invalid in list: $arg\n";
        exit($EXIT_ARG_PARSING);
      }
    }
  }
  else {
    print STDERR "Unrecognized switch: $arg\n";
  }
}

# Use default when value is not specified.
$opt{'verbosity'} = 1 if !defined $opt{'verbosity'};
$opt{'minPercentage'} = 0 if !defined $opt{'minPercentage'};

print STDERR "Payoff Matrix $VERSTR\n" if $opt{'showver'};

if ($opt{'dirnames'}) {
  foreach my $dirname ( @{$opt{'dirnames'}} ) {
    if (! -d $dirname) {
      print STDERR "Nonexistant folder: $dirname (skipping)\n"; next;
    }
    opendir DH, $dirname or die "Unable to open directory: $dirname";
    my @dfnames = grep { -f "$dirname/$_" } readdir(DH);
    closedir(DH);
    foreach my $fname (@dfnames) { push @fnames, "$dirname\\$fname"; }
  }
}

if (scalar(@fnames) == 0) {
  printf STDERR "No game files to process.\n"; exit($EXIT_SUCCESS);
}

if ( (defined $opt{'img1fname'} && $opt{'img1fname'} !~ /\.png/i) ||
     (defined $opt{'img2fname'} && $opt{'img2fname'} !~ /\.png/i) ) {
  print STDERR "Note: Image(s) are always written in PNG format regardless of file extension.\n\n";       
}

# Use masterpoint files at start and end of the date range for the game
# files examined if -mp is specified without an argument.
if ( $opt{'mpcols'} && !exists $opt{'mpdates'} ) {
  findMPdates(\@fnames);
  if ( $opt{'verbosity'} ) {
    my @yymm = sort keys $opt{'mpdates'};
    if ( scalar(@yymm) == 2 ) {
      print STDERR "Masterpoint data will be attached for $yymm[0] (start date) and $yymm[1] (end date).\n";
    }
    else {
      print STDERR "Masterpoint data will be attached for $yymm[0] (all games files are in one month).\n";
    }
  } 
}


# Check that all masterpoint databases needed have been downloaded already
# or can be downloaded.
if ( $opt{'mpcols'} ) {
  foreach my $yymm (sort keys $opt{'mpdates'}) {
    ($errmsg, undef) = checkMPdb(substr($yymm,0,2), substr($yymm,2,2));
    if ($errmsg) {
      print STDERR "Processing aborted because Masterpoint Database for $yymm is not available.\n";
      print STDERR "$errmsg\n";
      print STDERR "Note: Earliest database is 0905 (May 2009) and a few later months are missing.\n";
      print STDERR "Try specifying the months explicity, e.g. -mp 0905,1308\n";
      exit($EXIT_NEED_MP_DATABASE);
    }
  }
}

# $PL contains player data/statistics, $PR contains pair data/statistics.
# $PM contains payoff matrix data/statistics. $PNN maps player names to
# player numbers - allows in part for handling players who do not have
# an ACBL player number.
my ($pl, $pr, $pm, $pnn, $mp); my $pnnIndex = 0; my $upperLimits;
my $ngames = 0; my $highest_monthnum; my $nonPairSections = 0;

# Loop through and process all relevant ACBLscore game files.
foreach my $fname (@fnames) {
  if ($fname !~ /\.ac(m|a|e|l|b)$/i) { next; }
  (my $lastchar = substr($fname,-1)) =~ tr /a-z/A-Z/;
  if ( $opt{'sessions'} && index($opt{'sessions'}, $lastchar) == -1) { next; }

  (my $bname = $fname) =~ s/.*[\\\/]//;
  my $year = substr($bname,0,2);
  if ($year !~ /\d\d/) { $year = undef; }
  elsif ($year >= 80) { $year += 1900; }
  else { $year += 2000; }
  
  if ( $opt{'years'} && !exists $opt{'years'}{$year} ) { next; }
    
  if ( $opt{'days'} ) {
    if (!defined $year) { next; }
    my $dow = Date::DayOfWeek::dayofweek( substr($bname,4,2), substr($bname,2,2), $year );
    if (! $opt{'days'}[$dow] ) { next; }
  }
  my $monthnum = 0;
  if ($bname =~ /^\d{4}/) { $monthnum = $year * 12 + substr($bname,2,2); }
  $highest_monthnum = $monthnum if !defined $highest_monthnum || $monthnum > $highest_monthnum;
  
  my $errmsg = parseACBLgamefile($fname);
  if ($errmsg) { print STDERR $errmsg, "\n"; next; }
  $ngames++; 
}

my $nplayers1 = scalar(keys %$pl);
my $npairs1   = scalar(keys %$pr);
my $nmatrix1  = scalar(keys %$pm);

# Lookup masterpoint holdings for all players if -mp switch is used.
if ( $opt{'mpcols'} ) {
  my @pnums = keys %$pl;

  foreach my $yymm (sort keys $opt{'mpdates'}) {
    ($errmsg, my $pmp) = lookupMP($yymm, \@pnums);
    if ($errmsg) {
      print STDERR "Masterpoint lookup failure for $yymm\n$errmsg\n";
      exit($EXIT_BAD_MP_DATABASE);
    }
    else {
      $mp->{$yymm} = $pmp;
    }
  }
}

# If -ms option is specified, trim $PR to pairs meeting minimum number of
# sessions and $PM to interaction between pairs meeting the criterion. 
($pr, $pm) = trim($pr, $pm) if defined $opt{'minSessions'};

if ( defined $opt{'minPlayerSessions'} ) {
  foreach my $k (keys %$pl) { delete $pl->{$k} if $pl->{$k}{'cnt'} < $opt{'minPlayerSessions'}; }
}

$pl = meanstd($pl);
$pr = meanstd($pr);
$pm = meanstd($pm);

# Assign rank index (1..N) to each partnership.
ixAssign();

if ( defined $opt{'plfname'} ) {
  $errmsg = dumpPlayers($opt{'plfname'}, $pl);
  print STDERR "Player dump failed: $errmsg\n" if $errmsg;
} 
if ( defined $opt{'prfname'} ) {
  $errmsg = dumpPartnerships($opt{'prfname'}, $pr, $pl);
  print STDERR "Partnership dump failed: $errmsg\n" if $errmsg;
}
if ( defined $opt{'pmfname'} ) {
  $errmsg = dumpPayoffMatrix($opt{'pmfname'}, $pm, $pr);
   print STDERR "Payoff matrix dump failed: $errmsg\n" if $errmsg;
}
if ( defined $opt{'jsfname'} ) {
  $errmsg = dumpJSON($opt{'jsfname'});
   print STDERR "JSON dump failed: $errmsg\n" if $errmsg;
}

if ( $opt{'verbosity'} ) {
  print STDERR "Game files processed: $ngames\n";
  
  if ($ngames > 0) {
    # Note: Some pairs can not interact because they have a player in common. The
    # percentage calculation does not account for this, though it does account for
    # the fact that a partnership can not interact with itself.
    my $maxInteractions1 = $npairs1 * ($npairs1 - 1) / 2;
    print STDERR  "Total unique players: ", $nplayers1, "\n";
    print STDERR  "Total partnerships across all files and sessions: ", $npairs1, "\n";
    printf STDERR "Total interacting pairs of partnerships: %d  (%0.2f%%)\n",
      $nmatrix1, $npairs1 ? 100 * $nmatrix1 / $maxInteractions1 : 0;
      
    if (defined $opt{'minSessions'}) {
      print STDERR "\nAfter imposing $opt{'minSessions'} session minimum cut\n";
      
      my $nplayers2 = scalar(keys %$pl);
      my $npairs2   = scalar(keys %$pr);
      my $nmatrix2  = scalar(keys %$pm);
      
      my $maxInteractions2 = $npairs2 * ($npairs2 - 1) / 2;
      print STDERR  "Total unique players: ", $nplayers2, "\n";
      print STDERR  "Total partnerships across all files and sessions: ", $npairs2, "\n";
      printf STDERR "Total interacting pairs of partnerships: %d  (%0.2f%%)\n",
      $nmatrix2, $npairs2 ? 100 * $nmatrix2 / $maxInteractions2 : 0;
    }
  }
  
  if ($nonPairSections > 0) {
    print STDERR "Total non-pair sections skipped: $nonPairSections\n";
  }
}

if ( $opt{'dumpUpperLimits'} ) {
  print STDERR "\nUpper Limit  Count  (Number of games encountered with each upper limit)\n";
  for my $mplimit (sort {$a <=> $b} keys %$upperLimits) {
    printf STDERR "%8d MP %6d\n", $mplimit, $upperLimits->{$mplimit} if $mplimit;
  }
  printf STDERR "       Open %6d\n", $upperLimits->{'0'} if exists $upperLimits->{'0'};
}

makeImages() if scalar(keys %$pr) && (defined $opt{'img1fname'} || defined $opt{'img2fname'});

exit($EXIT_SUCCESS);


sub findMPdates {
  die "function requires 1 argument." if scalar(@_) != 1;
  my ($fnames) = @_;
  
  my (@bnames, $yy, $mm, $year);

  # The ACBL usually releases a new update of the Club Database on the 7th
  # of each month. But allow a week of slop but it may not yet be archived on
  # the La Jolla Bridge website.
  (undef, undef, undef, my $mday, my $avail_mon, my $avail_year) = localtime(time);
  # localtime() returns 0-based month and year since 1900. Convert to since 2000
  # because all masterpoint database available are after 2000.
  $avail_year -= 100;
  $avail_mon++ if $mday >= 15;
  if ($avail_mon == 0) { $avail_year--; $avail_mon = 12; }

  # It would be best to look at the dates for each event in each game file but for
  # now we will just rely on the YYMMDD style naming of ordinary club files.
  foreach my $fname (@$fnames) {
    (my $bname = $fname) =~ s/.*[\\\/]//;
    next if $bname !~ /^\d{6}\.ac(m|a|e|l|b)$/i;
    
    $year = substr($bname,0,2);
    if ($year !~ /\d\d/) { $year = undef; }
    elsif ($year >= 80) { $year += 1900; }
    else { $year += 2000; }
    
    next if $opt{'years'} && !exists $opt{'years'}{$year};
      
    if ( $opt{'days'} ) {
      if (!defined $year) { next; }
      my $dow = dayofweek( substr($bname,4,2), substr($bname,2,2), $year );
      next if ! $opt{'days'}[$dow];
    }

    push @bnames, $bname;
  }
  @bnames = sort(@bnames);
  $yy = substr($bnames[0], 0, 2);
  $mm = substr($bnames[0], 2, 2);
  
  if (12*$yy + $mm > 12*$avail_year + $avail_mon) {
    printf STDERR "Masterpoint database for %d-%02d is not available yet.\n",
      2000 + $yy, $mm;
    printf STDERR "Will try to use %d-%02d masterpoint database instead.\n",
      2000 + $avail_year, $avail_mon;
    $yy = $avail_year; $mm = $avail_mon; 
  }
  
  $opt{'mpdates'}{"$yy$mm"} = undef;
  
  $yy = substr($bnames[$#bnames], 0, 2);
  $mm = substr($bnames[$#bnames], 2, 2);  
  
  # If all game files are in the same month, do not create two YYMM files for the same month.
  return if exists $opt{'mpdates'}{"$yy$mm"};
  
  if (12*$yy + $mm > 12*$avail_year + $avail_mon) {
    printf STDERR "Masterpoint database for %d-%02d is not available yet.\n",
      2000 + $yy, $mm;
    printf STDERR "Will try to use %d-%02d masterpoint database instead.\n",
      2000 + $avail_year, $avail_mon;
    $yy = $avail_year; $mm = $avail_mon; 
  }
  
  $opt{'mpdates'}{"$yy$mm"} = undef;
}


sub parseACBLgamefile {
  # Parse binary ACBLscore output. This format isn't documented but I have partly
  # reverse engineered it for pairs games.
  
  die "function requires 1 argument." if scalar(@_) != 1;
  my ($fname) = @_;
  
  (my $bname = $fname) =~ s/.*[\\\/]//;
  
  my ($data, $errmsg, $pnum);
  # Safely isolate $/ change.
  {
    local $/; my $fh;
    return("Unable to open/read ACBLscore gamefile: $fname") if !open($fh, '<', $fname);
    binmode($fh);
    $data = <$fh>;
    close($fh);
  }
  
  # Check for the 'AC3' magic bytes that seem to appear at the start of all ACBLscore
  # game files.
  if (substr($data,0,6) ne "\x12\x0a\x03AC3") {
    return("Does not appear to be an ACBLscore game file: $fname");
  }
  
  print STDERR "Processing game file: $fname\n" if $opt{'verbosity'} > 1;
  
  my @sections;
  # List of sections begins at offset 0x13e and each section is 22 bytes long.
  my $ix = 0x13e;
  while ( ord(substr($data,$ix,1)) ) {
    my ($eventnum, $zLetter, $pDetails, $pBoards, $mp100, $secnum, $nplayed1, $nplayed2) =
      unpack('CC x2 VVv x C x3 CC', substr($data,$ix,22) );
    my $pEvent = unpack('V', substr($data,14+4*$eventnum));
    my $mpLimit = unpack('v', substr($data,$pEvent+0xf4,2));
    push @sections, {'eventnum' => $eventnum, 'letter' => substr($data,$ix+2,$zLetter),
      'pDetails' => $pDetails, 'pBoards' => $pBoards, 'mp100' => $mp100, 'secnum' => $secnum,
      'nplayed1' => $nplayed1, 'nplayed2' => $nplayed2, 'mpLimit', $mpLimit };
    $upperLimits->{$mpLimit}++;  
    $ix += 22;
  }
  
  if ($opt{'verbosity'} > 1) {
    print STDERR "List of sections\n";
    for my $i (0..$#sections) {
      my $sc = $sections[$i];
      my $mpStr = $sc->{'mpLimit'} ? $sc->{'mpLimit'} : 'Open';
      printf STDERR "%2s %2d %2d %7.2f %2d %2d %4s  Section Details at 0x%04x  Board Index at 0x%04x\n", 
        $sc->{'letter'}, $sc->{'eventnum'}, $sc->{'secnum'}, $sc->{'mp100'}, $sc->{'nplayed1'},
        $sc->{'nplayed2'}, $mpStr, $sc->{'pDetails'}, $sc->{'pBoards'};
    }
    print STDERR "\n";
  }
  
  # Process each section.
  for my $i (0..$#sections) {
    my $sc = $sections[$i];
    # Skip section if it is not a pairs section.
    if ( !$sc->{'pBoards'} ) { $nonPairSections++; next; }
    
    # Skip section if it does not have a desired masterpoint upper limit.
    next if defined $opt{'ul'} && !exists $opt{'ul'}{$sc->{'mpLimit'}};
    
    # Skip section if it belongs to an event number which is to be skipped.
    next if defined $opt{'eidx'} && !$opt{'eidx'}{$sc->{'eventnum'}};
    
    # $PS (as in "Pairs in Section")
    my ($errmsg, $isHowell, $ps) = parseSectionPairs($data, $sc->{'pDetails'}, $bname, $sc->{'letter'});
    if ($errmsg) {
      print STDERR "Problem with section $sc->{'letter'} of $bname: $errmsg (skipping)\n";
      next;
    }
 
    # Create unique identifier for each partnership by combining player numbers.
    # Also update parternship and player statistics
    foreach my $j (0..$#$ps) {
      my ($pnum1, $pnum2, $pair_id, $fl, $pct2); 
      while ( my($pairnum, $p) = each $ps->[$j] ) {
        # Assign pair_id of the form 'pnum1+pnum2'
        $pnum1 = $p->{'pnum'}[0]; $pnum2 = $p->{'pnum'}[1];
        $pair_id = $pnum1 < $pnum2 ? $pnum1 . '+' . $pnum2 : $pnum2 . '+' . $pnum1;
        $p->{'pair_id'} = $pair_id;
        
        # Update pair statistics.
        if ( !defined $pr->{$pair_id} ) {
          $fl = $pnum1 < $pnum2 ? 0 : 1;
          $pr->{$pair_id} = {
            'pnum1' => $p->{'pnum'}[$fl], 'pnum2' => $p->{'pnum'}[1-$fl],
            'fullname1' => $p->{'fname'}[$fl] . ' ' . $p->{'lname'}[$fl],
            'fullname2' => $p->{'fname'}[1-$fl] . ' ' . $p->{'lname'}[1-$fl],
            'mpsum' => 0, 'cnt' => 0, 'pctsum' => 0, 'pctsum2' => 0,
            'pctmax' => 0, 'slamcnts' => [0, 0, 0, 0], 'rulecnt' => 0
          };
        }
        $pct2 = $p->{'pct'} * $p->{'pct'};
        $pr->{$pair_id}{'cnt'}++;
        $pr->{$pair_id}{'pctsum'}  += $p->{'pct'};
        $pr->{$pair_id}{'pctsum2'} += $pct2;
        $pr->{$pair_id}{'mpsum'}   += $p->{'mp'};
        $pr->{$pair_id}{'pctmax'} = $p->{'pct'} if $pr->{$pair_id}{'pctmax'} < $p->{'pct'};
      
        # Update player statistics.
        for my $k (0..1) {
          if ( !defined $pl->{$p->{'pnum'}[$k]} ) {
            $pl->{$p->{'pnum'}[$k]} = {
              'fname' => $p->{'fname'}[$k], 'lname' => $p->{'lname'}[$k],
              'fullname' => $p->{'fname'}[$k] . ' ' . $p->{'lname'}[$k],
              'mpsum' => 0, 'cnt' => 0, 'pctsum' => 0, 'pctsum2' => 0,
              'pctmax' => 0
            };
          }
          $pnum = $p->{'pnum'}[$k];
          $pl->{$pnum}{'cnt'}++;
          $pl->{$pnum}{'pctsum'}  += $p->{'pct'};
          $pl->{$pnum}{'pctsum2'} += $pct2;
          $pl->{$pnum}{'mpsum'}   += $p->{'mp'};
          $pl->{$pnum}{'pctmax'} = $p->{'pct'} if $pl->{$pnum}{'pctmax'} < $p->{'pct'};
        }
      }
    }
    
    # Parse board results, updating $PM payoff matrix
    parseSectionBoards($data, $sc->{'pBoards'}, $ps, $isHowell, $bname, $sc->{'letter'});
  }
  
  return undef;
}

sub parseSectionPairs {
  # Parse information about pairs in a section.
  die "function requires 4 arguments." if scalar(@_) != 4;
  my ($data, $ix, $bname, $secLetter) = @_;
  
  my $PAIR_ENTRY_LEN = 0x192;
  
  my @ps; 
  my ($len, $iy);

  # The Section Details Table contains at least two, though possibly up to
  # four, pointers to Pair Index Tables. The two Pair Index Tables list
  # which pairs are seated N-S and E-W respectively at each table at round
  # one. Note: It is important to use the Pair Index Tables because although
  # the Pair entries are usually consecutive, this is not always true. One
  # exception occurs when a pair is added to a full Mitchell movement after
  # one round has been entered. (See 121125A U526 unit game, Pair 17 N-S in
  # section A.)
  
  my @pPairIndexTables = unpack('V2', substr($data,$ix+4,8));
  my @unk = unpack('V2', substr($data,$ix+12,8));
  if ( ($unk[0] || $unk[1]) and $opt{'verbosity'} ) {
    printf STDERR "Warning: Spare values in the Section Details Table that might be used as\n" .
      "additional Pair Table pointers are not both zero (0x%04x and 0x%04x). This\n" .
      "may indicate a movement that is not handled properly by this program.\n\n", 
      $unk[0], $unk[1];
  }
  
  my $pMovement = unpack('V', substr($data,$ix+20,4));
  my ($nTables, $nRounds, @tables, $tablenum);
  
  # Use the Movement table to figure out which pair number was seated N-S
  # and E-W at each table at round one. For Mitchell movements this is an
  # unnecessary extra step but it is the general technique. For Howell
  # movements it is essential.
  ($nTables, $nRounds) = unpack('CC', substr($data, $pMovement+4, 2));
  if ( $opt{'verbosity'} > 3 ) {
    printf STDERR "%d tables and %d rounds in Movement Table at 0x%04x\n",
      $nTables, $nRounds, $pMovement;
  }
  
  # The Movement Table has a 7 byte header and 75 bytes per table. The 75 bytes
  # represent up to 25 rounds where each group of three bytes gives the N-S pair
  # number at the table, the E-W pair number at the table, and the lowest board
  # number at the table for the round.
  $tables[0] = [ unpack("(C x74)$nTables",   substr($data, $pMovement+7, $nTables * 75)) ];
  $tables[1] = [ unpack("(x C x73)$nTables", substr($data, $pMovement+7, $nTables * 75)) ];

  # For a Howell movement, the pair numbers for the pairs seated N-S and E-W at
  # round 1 will not overlap.
  my $isHowell = 1;
  my @allpairnums = sort( ( @{$tables[0]}, @{$tables[1]} ) );
  for (my $i=0; $i<$#allpairnums-1; $i++) {
    if ( $allpairnums[$i] == $allpairnums[$i+1] ) { $isHowell = 0; last; }
  }

  if ( $opt{'verbosity'} > 3 ) {
    printf STDERR "\n %s Table:", $isHowell ? '  (Howell)' : '(Mitchell)';
    for my $i (1..$nTables) { printf STDERR " %2d", $i; }
    printf STDERR "\n", ;
    for my $i (0..1) {
      print STDERR "Round 1 ", $i ? "E-W" : "N-S", " pairs:";
      foreach my $pairnum (@{$tables[$i]}) { printf STDERR " %2d", $pairnum; }
      if ($i) { print STDERR "\n\n"; } else { print STDERR "\n"; } 
    }
  }
  
  
  my ($pairnum, $tot, $pct, $mp, $SA, $OA, $SB, $OB, $SC, $OC,
      @zlname, @zfname, @zcity, @zstate, @zpnum);
  my (@lname, @fname, @city, @state, @pnum);  
  
  my $nPairsRecorded = 0;
  for my $idir (0..1) {
    my $psIndex = $isHowell ? 0 : $idir;
    my $ip = $pPairIndexTables[$idir];
    if ($ip == 0) {
      printf STDERR "Warning: Pair Index Table %d seems to be missing.\n", $idir+1;
    };
    
    # This number includes phantom pairs.
    my $nPairs = unpack('C', substr($data, $ip+6, 1));
    if ( $opt{'verbosity'} > 3 ) {
      printf STDERR "%d pairs in Pair Index Table at 0x%04x\n", $nPairs, $ip;    
    }
    
    # Skip to first Pair Index Entry and advance 4 more bytes to location of pointer.
    $ip += 24;
    for (my $k=0; $k < $nPairs; $k++) {
      $ix = unpack('V', substr($data,$ip,4));
      $ip += 8;
      
      $len = unpack('v', substr($data,$ix,2));
      if ($len != $PAIR_ENTRY_LEN) { 
        printf STDERR "Pair Table Entry %d of section %s does not have expected " . 
          "length (%d) in file\n$bname. (Fatal error)\n", $k+1, $secLetter, $PAIR_ENTRY_LEN;
        exit;
      }

      # Each pair record has 162 bytes of pair data (beyond the record length) and two player
      # fields of 120 bytes = 402 bytes total (0x192).
      ($tablenum, $tot, $pct, $mp, $SA, $OA, $SB, $OB, $SC, $OC,
       $zlname[0], $zfname[0], $zcity[0], $zstate[0], $zpnum[0],
       $zlname[1], $zfname[1], $zcity[1], $zstate[1], $zpnum[1]) =
      unpack('x2 C x9 v x14 v x46 v x16 (C x3 C x15)3 x10 (C x16 C x16 C x16 C x2 C x7 x58)2',
        substr($data, $ix, $PAIR_ENTRY_LEN + 2) );
      $tot /= 100; $mp /= 100; $pct /= 100;
      
      # Convert table number to pair number using the Movement table data.
      if (!$isHowell && $tablenum > $nTables) {
        # This case is seen in 120427.ACA La Jolla unit game. This is a bit of kludge because
        # I don't know exactly how this case is treated in ACBLscore.
        print STDERR "A pair seems to be have been added to table $secLetter$tablenum of a Mitchell movement\n" .
          "after the first round in game file $bname. Will assume pair number\n" . 
          "is equal to the table number.\n\n" if $opt{'verbosity'};
        $pairnum = $tablenum;
      }
      else {
        $pairnum = $tables[$idir][$tablenum-1];        
      }
      
      # Check to avoid phantom pairs
      if ($zfname[0] != 0 || $zlname[0] != 0 || $zfname[1] != 0 || $zlname[1] != 0) {
        # Grab strings, canonicalize life masters and assign artificial player numbers to non-ACBL members.
        for my $i (0..1) {
          $nPairsRecorded++;
          
          $iy = $ix + ($i ? 0x11c : 0xa4);
          $lname[$i] = substr($data, $iy+1,  $zlname[$i]);
          $fname[$i] = substr($data, $iy+18, $zfname[$i]);
          $city [$i] = substr($data, $iy+35, $zcity [$i]);
          $state[$i] = substr($data, $iy+52, $zstate[$i]);
          $pnum [$i] = substr($data, $iy+55, $zpnum [$i]);        
    
          if ( $pnum[$i] =~ /^[1-9J-R][0-9]{6}$/ ) {
            # Canonicalize life masters.
            $pnum[$i] =~ tr/J-R/1-9/;            
          }
          else {
            if ($pnum[$i] ne 'NM' && $zpnum[$i] > 0) {
              # 'NM' is sometimes used for 'Non Member'. Anything else is weird. The 090219.ACE
              # Soledad Club file has '434900' for Ed Kennedy and the HTML report says 'BAD #'
              # underneath the player number. Maybe the ACBLscore input validation was not as
              # good back then. Warn the user but still assign an artificial player number.
              print STDERR "Bad player number: $pnum[$i] in file $bname. ",
                "(artificial number assigned)\n" if $opt{'verbosity'};
            }
            # Assign an artificial player number if we have not already done so.
            my $fullname = $fname[$i] . ' ' . $lname[$i];
            $pnn->{$fullname} = ++$pnnIndex if !defined $pnn->{$fullname};
            $pnum[$i] = $pnn->{$fullname};
          }

          if ($opt{'verbosity'} > 1) {
            printf STDERR "%2d %-30s %7d  %6.2f %5.2f %5s  %d %d %d  %d %d %d\n",
              $pairnum, $fname[$i] . ' ' . $lname[$i], $pnum[$i], 
              $tot, $pct, $mp ? sprintf('%5.2f', $mp) : '', $OA, $OB, $OC, $SA, $SB, $SC;
          }
          
        }
        print STDERR "\n" if $opt{'verbosity'} > 1;
        
        $ps[$psIndex]{$pairnum} = {'tot' => $tot, 'pct' => $pct, 'mp' => $mp, 'SA' => $SA, 'OA' => $OA,
          'SB' => $SB, 'OB' => $OB, 'SC' => $SC, 'OC' => $OC, 'lname' => [@lname], 'fname' => [@fname],
          'city' => [@city], 'state' => [@state], 'pnum' => [@pnum]};
      }      

    }
  }
  
  # This error can occur if we are processing an event that was setup names were never entered.
  my $errmsg = $nPairsRecorded ? undef : "No pairs found";
  return ($errmsg, $isHowell, \@ps);
}


sub parseSectionBoards {
  # Parse information about boards in a section.
  die "function requires 6 arguments." if scalar(@_) != 6;
  my ($data, $ix, $ps, $isHowell, $bname, $secLetter) = @_;
  
  my $verbose = $opt{'verbosity'} > 2;
  my $len = unpack('v', substr($data,$ix,2));
  my $nboards = unpack('C', substr($data,$ix+4));  

  # Note: Some slam scores match non-slam scores. Most of these scores are
  # very unlikely results. Moreover, accounting for vulnerability, knocks
  # out the collisions in square brackets.
  #
  # 920 is also [4m**+0 V], 4m**+1 NV, 1M**+2 NV, 4N**+0 NV.
  # 940 is also [1m*+4 V]. 980 is also [1N*+4 V, 2m*+5 V]. 990 is also 2N*+5 NV
  # [4M*+1 V]. 1010 is also [4M*+1 V]. 1090 is also [2N*+2 V]. 1190 is also
  # [4M*+2 V]. 1210 is also [4M*+2 V]. 1230 is also 1m*+5 NV. 1310 is also
  # [4m*+3 V]. 1330 is also [3M*+3 V]. 1380 is also [1N*+6 V]. 1390 is also
  # 4H*+3 V. 1430 is also 1m**+3 V and [1m**+6 NV]. 1440 is also [2M**+4 NV].
  # 1460 is also 2M*+4 V. 1520 is also [1M**+2 V]. 1680 is also 2N**+2 V.
  # 1830 is also 1m**+4 V. 1880 is also 4M**+2 V. 1960 is also [1N**+3 V].
  # 2230 is also 1m**+5 V. 2280 is also [4M**+3 V].
   
  # Slam score hashes. Key is slam score. Value is slam type: minor (0), major (1), NT (2).
  my %slamNV = (
    '920'  => 0, '940'  => 0, '1090' => 0, '1190' => 0, '1380' => 0, '1580' => 0,
    '980'  => 1, '1010' => 1, '1210' => 1, '1310' => 1, '1620' => 1, '1820' => 1,
    '990'  => 2, '1020' => 2, '1230' => 2, '1330' => 2, '1660' => 2, '1860' => 2,
    '1440' => 0, '1630' => 0, '1960' => 0, '1510' => 1, '1770' => 1, '2240' => 1,
    '1520' => 2, '1790' => 2, '2280' => 2
  );
  my %slamVul = (
    '1370' => 0, '1390' => 0, '1540' => 0, '1740' => 0, '1830' => 0, '2230' => 0,
    '1430' => 1, '1460' => 1, '1660' => 1, '1860' => 1, '2070' => 1, '2470' => 1,
    '1440' => 2, '1470' => 2, '1680' => 2, '1880' => 2, '2110' => 2, '2510' => 2,
    '2140' => 0, '2330' => 0, '2660' => 0, '2210' => 1, '2470' => 1, '2940' => 1,
    '2220' => 2, '2490' => 2, '2980' => 2
  );
  
  print STDERR "Number of boards: $nboards\n" if $verbose;
  
  # 38 bytes gets to the first board index entry. File offset is four bytes in. 
  $ix += 42;
  
  my($iy, $blen, $bnum, $nresults, $mp_top, $slam_type);
  my($round, $ns_pair, $ew_pair, $ns_score, $ew_score, $ns_mp, $ew_mp);
  my($ns_pair_id, $ew_pair_id, $pct, $pair_id1, $pair_id2, $pair_pair_id);

  for (my $nb = 0; $nb < $nboards; $nb++) {
    # Each board has results for every round. If a board was not in play on a
    # given round of the movement or it met a phantom pair, special scores
    # indicate this situation.
    $iy = unpack('V', substr($data,$ix,4));
    ($blen, $bnum, $nresults) = unpack('vC x2 C', substr($data,$iy,6) );
    if ($nresults * 6 + 18 == $blen) {
      die "Inconsistent board table length.  blen: $blen  nresults: $nresults";
    }
    # Skip to the first result for the board.
    $iy += 6;
    
    print STDERR "\nBoard $bnum\n" if $verbose;
    for (my $nr = 0; $nr < $nresults; $nr++ ) {
      # Score unpacks as 1/10 of actual score, i.e. superfluous zero is dropped
      # The ACBLscore interface for entering scores does the same thing. The
      # matchpoint scores are 100 times the actual value. Note: we have both
      # N-S and E-W scores because director rulings can result in split scores.
      # Also the byte before the N-S pair number seems to always match the N-S
      # pair number for a Mitchell movement but is different for a Howell movement.
      ($round, $ns_pair, $ns_score, $ns_mp, $ew_pair, $ew_score, $ew_mp) =
        unpack('C x (C x s< v x2)2', substr($data,$iy,18) );
      $iy += 18;  
      
      if ($verbose) {
        printf STDERR "%2d  %2d vs %2d %5d %5d %4.2f %4.2f\n", $round, $ns_pair, $ew_pair,
          $ns_score * 10, $ew_score * 10, $ns_mp / 100, $ew_mp / 100;
      }
      
      # Board was not in play on this round.
      next if ($ns_pair == 0 || $ew_pair == 0);
      
      $ns_pair_id = $ps->[0]{$ns_pair}{'pair_id'};
      $ew_pair_id = $ps->[$isHowell ? 0 : 1]{$ew_pair}{'pair_id'};
      
      # Pair lookup can fail if the Pair Table for the section was not read correctly.
      # This can happen in weird cases, e.g. when a pair is added after a round is
      # already played such that the pairs in the Pair Table are not all contiguous.
      if (!defined $ns_pair_id || !defined $ew_pair_id) {
        print STDERR "Pair lookup failure: $bname $secLetter:", 
        defined $ns_pair_id ? '' : " N-S $ns_pair",
        defined $ew_pair_id ? '' : " E-W $ew_pair", "\n";
        next;
      }
           
      # Ignore split scores (they result from director rulings), Avg, Avg+, and Avg-
      # results. 7600 (down 13 redoubled vul), stored as 760, is the best/worst score
      # a pair can recieve. Values higher than this are special (950 => Not Played,
      # 999 => Placeholder (board not in played on round), 2040 => Ave-, 2050 => Ave,
      # 2060 => Ave+).
      if ( ($ns_score + $ew_score) != 0 || $ns_score > 760 || $ew_score > 760) {
        # 999 check is probably redundant with ($ns_pair == 0 || $ew_pair == 0) check
        if ($ns_score != 950 && $ns_score != 999) {
          # Keep track of which partnerships cause director rulings.
          $pr->{$ns_pair_id}{'rulecnt'}++; $pr->{$ew_pair_id}{'rulecnt'}++;
        }
        next;
      }

      # It doesn't make sense to compute this for every result but one must be careful
      # about using the number of times the board is play due to scaling when sections
      # are combined and boards are played an unequal number of times.
      $mp_top = ($ns_mp + $ew_mp) / 100;
      
      if ($ns_pair_id le $ew_pair_id) {
        $pair_id1 = $ns_pair_id; $pair_id2 = $ew_pair_id;
        $pct = $ns_mp / $mp_top;
      }
      else {
        $pair_id1 = $ew_pair_id; $pair_id2 = $ns_pair_id;
        $pct = $ew_mp / $mp_top;
      }
            
      $pair_pair_id = $pair_id1 . ':' . $pair_id2;
      if ( !defined $pm->{$pair_pair_id} ) {
        $pm->{$pair_pair_id} = { 'cnt' => 1, 'pctsum' => $pct, 'pctsum2' => $pct * $pct };
      }
      else {
        $pm->{$pair_pair_id}{'cnt'}++;
        $pm->{$pair_pair_id}{'pctsum'}  += $pct;
        $pm->{$pair_pair_id}{'pctsum2'} += $pct * $pct;
      }
      
      # Credit partnerships with any slams bid and made. No slam scores are divisible
      # by 100, so check for this immediately to avoid a hash lookup.
      if ( $ns_score >= 92 && $ns_score % 10 != 0 ) {
        $slam_type = $isVulNS[$bnum % 16] ? $slamVul{$ns_score * 10} : $slamNV{$ns_score * 10};
        if (defined $slam_type) {
          $pr->{$ns_pair_id}{'slamcnts'}[3]++;
          $pr->{$ns_pair_id}{'slamcnts'}[$slam_type]++;
        }
      }
      elsif ( $ew_score >= 92 && $ew_score % 10 != 0 ) {
        $slam_type = $isVulEW[$bnum % 16] ? $slamVul{$ew_score * 10} : $slamNV{$ew_score * 10};
        if (defined $slam_type) {
          $pr->{$ew_pair_id}{'slamcnts'}[3]++;
          $pr->{$ew_pair_id}{'slamcnts'}[$slam_type]++;
        }
      }
      
    }
    $ix += 8;
  } 


}

sub trim {
  die "function requires 2 arguments." if scalar(@_) != 2;
  my ($pr, $pm) = @_;
  
  my ($pair_id1, $pair_id2);
  foreach my $k (keys %$pr) { delete $pr->{$k} if $pr->{$k}{'cnt'} < $opt{'minSessions'}; }
  foreach my $k (keys %$pm) {
    ($pair_id1, $pair_id2) = split /\:/, $k;
    delete $pm->{$k} if !defined $pr->{$pair_id1} || !defined $pr->{$pair_id2};
  }
  
  return ($pr, $pm);
}

sub ixAssign {
  my $ix = 0;
  my @keys = sort {$pr->{$b}{'pct'} <=> $pr->{$a}{'pct'} } keys %$pr;
  foreach my $k (@keys) { $ix++; $pr->{$k}{'ix'} = $ix; }
}

sub dumpPlayers {
  # Dump player information sorted by decreasing average percentage.
  die "function requires 2 arguments." if scalar(@_) != 2;
  my ($fname, $pl) = @_;
  
  return "Unable to open/write: $fname" if ! open(my $fh, '>', $fname);

  my ($mpdate, @mpdates, $mpval);
  my @hrow = ('pnum', 'fname', 'lname', 'fullname', 'cnt', 'pct', 'meanerr', 'std', 
    'pctmax', 'mpsum', 'mpavg');
  if ( $opt{'mpcols'} ) {
    @mpdates = sort keys $opt{'mpdates'};
    foreach $mpdate (@mpdates) { push @hrow, 'mp' . $mpdate; }
  }
 
  print $fh join("\t", @hrow), "\n";
  
  my (@row, $pct, $meanerr, $std, $mpavg, $pctmax);
  my @keys = sort {$pl->{$b}{'pct'} <=> $pl->{$a}{'pct'} } keys %$pl;
  foreach my $k (@keys) {
    last if $pl->{$k}{'pct'} < $opt{'minPercentage'};
    
    $pct = sprintf('%.2f', $pl->{$k}{'pct'});
    $std = defined $pl->{$k}{'std'} ? sprintf('%.2f', $pl->{$k}{'std'}) : '';
    $meanerr = defined $pl->{$k}{'meanerr'} ? sprintf('%.2f', $pl->{$k}{'meanerr'}) : '';
    $mpavg = sprintf('%.2f', $pl->{$k}{'mpsum'} / $pl->{$k}{'cnt'});
    $pctmax = sprintf('%.2f', $pl->{$k}{'pctmax'});
    
    @row = ( $k, $pl->{$k}{'fname'}, $pl->{$k}{'lname'}, $pl->{$k}{'fullname'},
      $pl->{$k}{'cnt'}, $pct, $meanerr, $std, $pctmax, $pl->{$k}{'mpsum'}, $mpavg );
    if ( $opt{'mpcols'} ) {
      foreach $mpdate (@mpdates) {
        $mpval = $mp->{$mpdate}{$k};
        push @row, defined $mpval ? sprintf('%.2f', $mpval) : '';
      }
    }
    print $fh join("\t", @row), "\n";
  }
  
  close($fh);
  return undef;
}

sub dumpPartnerships {
  # Dump partnership information sorted by decreasing average percentage.
  die "function requires 3 arguments." if scalar(@_) != 3;
  my ($fname, $pr, $pl) = @_;
  
  return "Unable to open/write: $fname" if ! open(my $fh, '>', $fname);

  my ($mpdate, @mpdates, $mpval, $pnum);  
  my @hrow = ('ix', 'pnum1', 'pnum2', 'fullname1', 'fullname2', 'cnt', 'pct',
    'meanerr', 'std', 'pctmax', 'mpsum', 'mpavg', 'slamcnt', 'slamavg', 'rulecnt');
  push @hrow, 'slamNT', 'slamMJ', 'slamMN' if $opt{'dumpSlamDetails'};
  if ( $opt{'mpcols'} ) {
    @mpdates = sort keys $opt{'mpdates'};
    foreach $mpdate (@mpdates) { push @hrow, 'mp' . $mpdate; }
    foreach $mpdate (@mpdates) { push @hrow, 'mp' . $mpdate; }
  }  
  print $fh join("\t", @hrow), "\n";
  
  my (@row, $pnum1, $pnum2, $pct, $meanerr, $std, $pctmax, $mpavg, $slamavg);
  my @keys = sort {$pr->{$a}{'ix'} <=> $pr->{$b}{'ix'} } keys %$pr;
  foreach my $k (@keys) {
    last if $pr->{$k}{'pct'} < $opt{'minPercentage'};
    
    ($pnum1, $pnum2) = split /\+/, $k;
    $pct = sprintf('%.2f', $pr->{$k}{'pct'});
    $std = defined $pr->{$k}{'std'} ? sprintf('%.2f', $pr->{$k}{'std'}) : '';
    $meanerr = defined $pr->{$k}{'meanerr'} ? sprintf('%.2f', $pr->{$k}{'meanerr'}) : '';
    $pctmax = sprintf('%.2f', $pr->{$k}{'pctmax'});
    $mpavg = sprintf('%.2f', $pr->{$k}{'mpsum'} / $pr->{$k}{'cnt'});
    $slamavg = sprintf('%.2f', $pr->{$k}{'slamcnts'}[3] / $pr->{$k}{'cnt'});
        
    @row = ($pr->{$k}{'ix'}, $pnum1, $pnum2, $pl->{$pnum1}{'fullname'}, $pl->{$pnum2}{'fullname'},
      $pr->{$k}{'cnt'}, $pct, $meanerr, $std, $pctmax, $pr->{$k}{'mpsum'}, $mpavg,
      $pr->{$k}{'slamcnts'}[3], $slamavg, $pr->{$k}{'rulecnt'} );
    push @row, $pr->{$k}{'slamcnts'}[0], $pr->{$k}{'slamcnts'}[1], $pr->{$k}{'slamcnts'}[2]
      if $opt{'dumpSlamDetails'};
    if ( $opt{'mpcols'} ) {
      foreach $pnum ($pnum1, $pnum2) {
        foreach $mpdate (@mpdates) {
          $mpval = $mp->{$mpdate}{$pnum};
          push @row, defined $mpval ? sprintf('%.2f', $mpval) : '';
        }
      }
    }      
    print $fh join("\t", @row), "\n";
  }
  
  close($fh);
  return undef;
}

sub dumpPayoffMatrix {
  # Dump payoff matrix sorted by decreasing disparity between partnerships.
  die "function requires 3 arguments." if scalar(@_) != 3;
  my ($fname, $pm, $pr) = @_;
  
  return "Unable to open/write: $fname" if ! open(my $fh, '>', $fname);

  my @hrow = ('ix1', 'ix2', 'pair_id1', 'pair_id2', 'pair_name1', 'pair_name2', 'cnt', 'pct', 
    'meanerr', 'std', 'disparity', 'relperf');
  print $fh join("\t", @hrow), "\n";
  
  my (@row, $pair_id1, $pair_id2, $pair_name1, $pair_name2, $pct, $std, $meanerr,
    $disparity, $relperf);
  my @keys = sort { abs(50 - $pm->{$b}{'pct'}) <=> abs(50 - $pm->{$a}{'pct'}) } keys %$pm;
  foreach my $k (@keys) {
    ($pair_id1, $pair_id2) = split /\:/, $k;
    $pair_name1 = $pr->{$pair_id1}{'fullname1'} . ' - ' . $pr->{$pair_id1}{'fullname2'};
    $pair_name2 = $pr->{$pair_id2}{'fullname1'} . ' - ' . $pr->{$pair_id2}{'fullname2'};
    
    $pct = sprintf('%.2f', $pm->{$k}{'pct'});
    $std = defined $pm->{$k}{'std'} ? sprintf('%.2f', $pm->{$k}{'std'}) : '';
    $meanerr = defined $pm->{$k}{'meanerr'} ? sprintf('%.2f', $pm->{$k}{'meanerr'}) : '';
    $disparity = sprintf('%.2f', 2 * abs(50 - $pm->{$k}{'pct'}) );
    $relperf = sprintf('%.2f', $pm->{$k}{'pct'} - 50 - $pr->{$pair_id1}{'pct'} + $pr->{$pair_id2}{'pct'});
    
    @row = ($pr->{$pair_id1}{'ix'}, $pr->{$pair_id2}{'ix'}, $pair_id1, $pair_id2,
      $pair_name1, $pair_name2, $pm->{$k}{'cnt'}, $pct, $meanerr, $std, $disparity, $relperf);
    print $fh join("\t", @row), "\n";
  }
  
  close($fh);
  return undef;
}

sub dumpJSON {
  die "function requires 1 argument." if scalar(@_) != 1;
  my ($fname) = @_;
  
  requirePackage('JSON', 'Dumping JSON output');
  
  # Pack data in a way that minimizes amount of output, i.e. as a bunch of arrays of the same
  # length instead of an array of hashes, for faster web page loading.
  my ($pnum1, $pnum2, $fd, $pj, $pair_id1, $pair_id2, $pct, $meanerr, $relperf);
  my $pairs = {'fullname1' => [], 'fullname2' => [], 'cnt' => [], 'pct' => [], 'std' => [], 'meanerr' => []};
  my @keys = sort {$pr->{$a}{'ix'} <=> $pr->{$b}{'ix'} } keys %$pr;
  foreach my $k (@keys) {
    ($pnum1, $pnum2) = split /\+/, $k;
    push @{$pairs->{'fullname1'}}, $pl->{$pnum1}{'fullname'};
    push @{$pairs->{'fullname2'}}, $pl->{$pnum2}{'fullname'};
    push @{$pairs->{'cnt'}}, $pr->{$k}{'cnt'};
    foreach my $fd ('pct', 'std', 'meanerr') {
      push @{$pairs->{$fd}}, int(100 * $pr->{$k}{$fd} + 0.5) / 100;
    }
  }
  
  # Dump Payoff Matrix as a sparse matrix.
  while (my($k, $v) = each %$pm) {
    ($pair_id1, $pair_id2) = split /\:/, $k;
    $pct = int(100 * $v->{'pct'} + 0.5) / 100;
    $meanerr = int(100 * $v->{'meanerr'} + 0.5) / 100;
    $relperf = $pm->{$k}{'pct'} -50 - $pr->{$pair_id1}{'pct'} + $pr->{$pair_id2}{'pct'};
    $relperf = int(100 * $relperf + 0.5) / 100;
    
    # Zero based pair numbers for convenience in Javascript code.  
    $pj->{ $pr->{$pair_id1}{'ix'}-1 }{ $pr->{$pair_id2}{'ix'}-1 } = 
      [ $v->{'cnt'}, $pct, $meanerr, $relperf ];
  }
  
  my $data = {'pairs' => $pairs, 'pm' => $pj};

  return "Unable to open/write: $fname" if ! open(my $fh, '>', $fname);
  
  (my $bname = $fname) =~ s/.*[\\\/]//;
  $bname =~ s/\.[^.]*$//;
  $bname =~ s/-payoff-matrix$//;
  
  printf $fh "payoffMatrix[\"%s\"] = ", $bname;
  print $fh JSON->new->encode($data);
  print $fh ';';

  close($fh);
  return undef;
}

sub meanstd {
  # Calculate average percentage, standard deviation, and error on average percentage.
  die "function requires 1 argument." if scalar(@_) != 1;
  my ($x) = @_;
  
  my $std2;
  while ( my ($key, $p) = each %$x ) {
    $p->{'pct'} = $p->{'pctsum'} / $p->{'cnt'};
    if ($p->{'cnt'} > 1) {
      $std2 = ($p->{'pctsum2'} - $p->{'pctsum'} * $p->{'pctsum'} / $p->{'cnt'}) / ($p->{'cnt'}-1)
        ;
      # If all data have the same value, roundoff error can cause the calculation
      # of the standard deviation to fail. Handle this but still fail if STD2 is too
      # negative since that would indicate a problem upstream.
  
      $std2 = 0 if $p->{'cnt'} > 1 && $std2 < 0 && $std2 > -1e-10;
      $p->{'std'} = sqrt($std2);
      $p->{'meanerr'} = $p->{'std'} / sqrt($p->{'cnt'});
    }
    else {
      $p->{'std'}     = undef;
      $p->{'meanerr'} = undef;
    }
 
    
  }
  return $x;
}

sub lookupMP {
  # Lookup masterpoint holdings for list of canonicalized player numbers.
  #
  # YYMM     - Masterpoint database to use
  # PPNUMS   - Arrayref of player numbers
  #
  # ERRMSG   - Error message (empty if success)
  # PMP      - Hashref of player masterpoints keyed by player number
  # MPFAIL   - Arrayref of masterpoint lookups that failed

  die "function requires 2 arguments." if scalar(@_) != 2;
  my ($yymm, $ppnums) = @_;

  die "PPNUMS must be an arrayref." if (ref($ppnums) ne 'ARRAY');
  
  my (%pmp, @mpfail);
  # Don't read database if there are no player numbers to lookup.
  if (scalar(@$ppnums) == 0) { return ('', %pmp, @mpfail); }

  my @pnums = sort {$a <=> $b} @$ppnums;
  
  my $MPdbname = appDir() . $yymm . 'D00MP.dat';
  open(my $fh, '<', $MPdbname) or return("Unable to open/read: $MPdbname ($!)");
  binmode($fh);

  # Check header.
  read($fh, my $id, 4);
  if ($id ne 'MPDB') { return "Not an ACBLmerge masterpoint database: $MPdbname"; }
  # Skip over year and month data.
  read($fh, my $yyyymm, 4);
  my ($db_year, $db_mon) = unpack('nn', $yyyymm);
  if ($db_year != substr($yymm,0,2) + 2000 || $db_mon != substr($yymm,2,2)) {
    return "Database data header is inconsistent with filename: $MPdbname";
  }

  local $/ = undef;
  my $data = <$fh>;
  close($fh);
  
  my $ix = 0; my $dlen = length($data);
  my $ar_pnum = shift(@pnums);
  my $db_pnum = unpack('N', substr($data, $ix, 4));
  
  # Perform sort merge lookup.
  while (1) {
    if ($db_pnum < $ar_pnum) {
      $ix += 8; if ($ix >= $dlen) { last; }
      $db_pnum = unpack('N', substr($data, $ix, 4));
    }
    elsif ($ar_pnum == $db_pnum) {
      $pmp{$db_pnum} = unpack('N', substr($data, $ix+4, 4)) / 100;
      $ix += 8; if ($ix >= $dlen) { last; }
      $db_pnum = unpack('N', substr($data, $ix, 4));
      $ar_pnum = shift(@pnums);
      if (!defined $ar_pnum) { last; }
    }
    else {
      push @mpfail, $ar_pnum;
      $ar_pnum = shift(@pnums);
      if (!defined $ar_pnum) { last; }
    }
  }
  
  push @mpfail, $ar_pnum if $ar_pnum;
  push @mpfail, @pnums;

  return ('', \%pmp, \@mpfail);
}

sub makeImages {
  requirePackage('GD', 'Creating Payoff Matrix image(s)');
  
  # Parameters for determining pixel size if use does not specify a size
  my $TARGET_SIZE = 600;
  my $MIN_CELL_SIZE = 6;
  # Default dynamic range used when displaying data (one side from center)
  my $DEFAULT_DYNAMIC_RANGE = 30;
  
  my @POS_COLOR = (255, 255, 0);
  my @NEG_COLOR = (0, 0, 255);
  my @blend1 = (0, 0, 0);
  my @blend2 = (0, 0, 0);
  
  my $dynRange1 = defined $opt{'dynRange1'} ? $opt{'dynRange1'} : $DEFAULT_DYNAMIC_RANGE;
  my $dynRange2 = defined $opt{'dynRange2'} ? $opt{'dynRange2'} : $DEFAULT_DYNAMIC_RANGE;
  my $lowConfCut = defined $opt{'lowconf'} ? $opt{'lowconf'} : 0;
  
  my $npairs = scalar(keys %$pr);
  my $px = $opt{'npixels'};
  if (!defined $px) {
    # Figure out how many pixels wide and high each cell in the image matrix should be
    $px = int($TARGET_SIZE / $npairs + 0.5);
    $px = $MIN_CELL_SIZE if $px < $MIN_CELL_SIZE;    
  }
  my $isize = $px * $npairs;
  
  # Initialize true color image(s). First is for the regular payoff matrix and the
  # second is for the over/underperform matrix.
  my ($im1, $im2, $im);
  $im1 = new GD::Image($isize, $isize, 1) if defined $opt{'img1fname'};
  $im2 = new GD::Image($isize, $isize, 1) if defined $opt{'img2fname'};
  
  my $no_data_color   = $im1->colorAllocate(255,255,255);
  my $cant_meet_color = $im1->colorAllocate(255,128,128);
  my $low_cnt_color   = $im1->colorAllocate(192,128,128);
  
  # Initialize images to missing data color (white).
  $im1->filledRectangle(0, 0, $isize-1, $isize-1, $no_data_color) if $im1;
  $im2->filledRectangle(0, 0, $isize-1, $isize-1, $no_data_color) if $im2;
  
  my ($pair_pair_id, $pair_id1, $pair_id2, $p, $relperf, $color);
  my $ix1 = 0; my $ix2 = 0; my $v = 0; my $frac = 0; my $i = 0; my $j = 0;
  my ($x1, $y1, $x2, $y2) = 0 x 4;
  while ( ($pair_pair_id, $p) = each %$pm ) {
    ($pair_id1, $pair_id2) = split /:/, $pair_pair_id;
    $ix1 = $pr->{$pair_id1}{'ix'};
    $ix2 = $pr->{$pair_id2}{'ix'};
    $x1 = $px * ($ix2-1); $y1 = $px * ($ix1 -1);
    $x2 = $px * $ix2 - 1; $y2 = $px * $ix1 - 1;
    
    if ($im1) {
      $v = ($p->{'pct'} - 50) / $dynRange1;
      if ($v < -1) { $v = -1; } elsif ($v > 1) { $v = 1; }
      $frac = ($v+1) / 2;
      for $i (0..2) {
        $blend1[$i] = $frac * $POS_COLOR[$i] + (1-$frac) * $NEG_COLOR[$i];
        $blend2[$i] = $frac * $NEG_COLOR[$i] + (1-$frac) * $POS_COLOR[$i];
      }
      
      $color = $im1->colorAllocate($blend1[0], $blend1[1], $blend1[2]);
      $im1->filledRectangle($x1, $y1, $x2 ,$y2 , $color);
      $color = $im1->colorAllocate($blend2[0], $blend2[1], $blend2[2]);
      $im1->filledRectangle($y1, $x1, $y2, $x2, $color);      
    }
    if ($im2) {
      $relperf = $p->{'pct'} -50 - $pr->{$pair_id1}{'pct'} + $pr->{$pair_id2}{'pct'};
      $v = $relperf / $dynRange2;
      if ($v < -1) { $v = -1; } elsif ($v > 1) { $v = 1; }
      $frac = ($v+1) / 2;
      for $i (0..2) {
        $blend1[$i] = $frac * $POS_COLOR[$i] + (1-$frac) * $NEG_COLOR[$i];
        $blend2[$i] = $frac * $NEG_COLOR[$i] + (1-$frac) * $POS_COLOR[$i];
      }
      
      $color = $im2->colorAllocate($blend1[0], $blend1[1], $blend1[2]);
      $im2->filledRectangle($x1, $y1, $x2 ,$y2 , $color);
      $color = $im2->colorAllocate($blend2[0], $blend2[1], $blend2[2]);
      $im2->filledRectangle($y1, $x1, $y2, $x2, $color);      
    }
    
    if ($p->{'cnt'} < $lowConfCut) {
      for ($i=2; $i<$px-2; $i++) {
        foreach $im ($im1, $im2) {
          next if !$im;
          $im->setPixel($x1 + $i, $y1 + $i, $low_cnt_color);
          $im->setPixel($x2 - $i, $y1 + $i, $low_cnt_color);
          $im->setPixel($y1 + $i, $x1 + $i, $low_cnt_color);
          $im->setPixel($y2 - $i, $x1 + $i, $low_cnt_color);
        }
      }
    }
    
  }
  
  # A partnership can never interact with itself. Flag the diagonal.
  for ($i=0; $i<$npairs; $i++) {
    $x1 = $px * $i; $x2 = $px * ($i+1) - 1;
    $im1->filledRectangle($x1, $x1, $x2, $x2, $cant_meet_color) if $im1;
    $im2->filledRectangle($x1, $x1, $x2, $x2, $cant_meet_color) if $im2;
  }
  
  # Figure out which pairs can never interact because they have one player in common.
  my %pnumix; my $ncommon = 0;
  while ( (undef, $p) = each %$pr ) {
    push @{$pnumix{$p->{'pnum1'}}}, $p->{'ix'};
    push @{$pnumix{$p->{'pnum2'}}}, $p->{'ix'};
  }
  while ( (undef, $p) = each %pnumix ) {
    $ncommon = scalar(@{$p});
    next if $ncommon == 1;
    for ($i=0; $i<$ncommon; $i++) {
      for ($j=0; $j<$i; $j++) {
        $ix1 = ${$p}[$i];
        $ix2 = ${$p}[$j];
        if ($im1) {
          $im1->filledRectangle($px * ($ix2-1), $px * ($ix1 -1), $px * $ix2 -1, $px * $ix1 - 1, $cant_meet_color);
          $im1->filledRectangle($px * ($ix1-1), $px * ($ix2 -1), $px * $ix1 -1, $px * $ix2 - 1, $cant_meet_color);          
        }
        if ($im2) {
          $im2->filledRectangle($px * ($ix2-1), $px * ($ix1 -1), $px * $ix2 -1, $px * $ix1 - 1, $cant_meet_color);
          $im2->filledRectangle($px * ($ix1-1), $px * ($ix2 -1), $px * $ix1 -1, $px * $ix2 - 1, $cant_meet_color);          
        }
      }
    }
  }
  
  # Write out image file(s).
  my $fh;
  if ($im1) {
    if (! open($fh, '>', $opt{'img1fname'})) {
      print STDERR "Unable to open/write: $opt{'img1fname'}\n";
    }
    else {
      binmode $fh; print $fh $im1->png; close($fh);    
    }
  }
  if ($im2) {
    if (! open($fh, '>', $opt{'img2fname'})) {
      print STDERR "Unable to open/write: $opt{'img2fname'}\n";
    }
    else {
      binmode $fh; print $fh $im2->png; close($fh);    
    }
  }

}

sub checkMPdb {
  die "function requires 2 arguments." if scalar(@_) != 2;
  my ($yy, $mm) = @_;

  if ($yy !~ /^\d+$/) { die "function requires integer year."; }
  if ($mm !~ /^\d+$/ || $mm < 1 || $mm > 12) { die "function requires month 1-12."; }
  
  my $MPdbname = appDir() . sprintf('%02d%02dD00MP.dat', $yy, $mm);
  if (! -f $MPdbname || -z $MPdbname) {
    $errmsg = fetchMPdb($MPdbname, $yy, $mm);
    if ($errmsg) {
      $errmsg = sprintf("Unable to fetch Masterpoint Database for %d-%02d (%s)\n",
        2000 + $yy, $mm, $errmsg);
      return($errmsg);
    }
  }
  
  open(my $fh, '<', $MPdbname) or return('Unable to open/read: $MPdbname ($!)');
  binmode($fh);

  # Check header.
  read($fh, my $id, 4);
  if ($id ne 'MPDB') { return "Not a masterpoint database: $MPdbname"; }
  read($fh, my $yyyymm, 4);
  close($fh);
  my ($db_year, $db_mon) = unpack('nn', $yyyymm);
  if ($db_year != $yy + 2000 || $db_mon != $mm) {
    return("Database data header is inconsistent with filename.");
  }
  
  return('', $MPdbname);
}


sub fetchMPdb {
  # Fetches masterpoint database for a given year and month. The masterpoint
  # database is derived from the ACBL Club Update database but previous months
  # are available from http://lajollabridge.com unlike the ACBL website which
  # only has the latest masterpoint database.
  if (scalar(@_) != 3) { die "function requires 3 arguments."; }
  my ($ofname, $yy, $mm) = @_;
  if ($yy !~ /^\d+$/) { die "function requires integer year."; }
  if ($mm !~ /^\d+$/ || $mm < 1 || $mm > 12) { die "function requires month 1-12."; }
  
  my $appdir = appDir();

  requirePackage('LWP::UserAgent', 'Fetching masterpoint database');

  my $URL = sprintf('http://lajollabridge.com/Software/MPdb/%02d%02dD00MP.dat', $yy, $mm);
  print STDERR "Fetching $yy$mm masterpoint database...\n" if $opt{'verbosity'}; 

  # Create a user agent object.
  my $ua = LWP::UserAgent->new;
  $ua->agent('Payoff-Matrix');
  $ua->timeout(10);

  # Create a request.
  my $req = HTTP::Request->new(GET => $URL);
  $req->header(Accept => "text/html, */*;q=0.1");

  # Pass request to the user agent and get a response back.
  my $res = $ua->request($req);

  # Check the outcome of the response
  if ($res->is_success) {
	  open(my $fh, '>', $ofname) or return "Can't open/write: $ofname";
	  binmode($fh);
	  print $fh $res->content;
	  close($fh);
	}
	else {
	  return $res->status_line;
	}
	
	return('');
}

sub appDir {
  # Returns the user's application directory for ACBLmerge.
  my $dname = defined $opt{'appdir'} ? $opt{'appdir'} :
    ($^O eq 'MSWin32') ? $ENV{'APPDATA'} . '\\Payoff-Matrix\\' : $ENV{'HOME'} . '/.Payoff-Matrix/';
  if (!-d $dname && !mkdir($dname)) {
    print STDERR "Unable to create per user application directory: $dname\n";
    exit($EXIT_MAKE_APPDIR_FAIL);
  }
  return($dname);
}

sub requirePackage {
  # FUNCTIONALTY - Optional description of software functionality requiring the package.
  my ($packageName, $functionality) = @_;

  eval "require $packageName";
  if ($@) {
    if ($^O ne 'MSWin32') {
      print STDERR <<"DONE";
$functionality
requires the $packageName Perl package.
Please install it. Automatic package installation is only provided for the
ActiveState distribution on Windows. If you have CPAN installed, the following
command might work:

  cpan -i $packageName  

DONE
      exit($EXIT_NEED_PERL_PACKAGE);
    }
    
    my $cmd = "ppm install $packageName";

    if (! $opt{'quiet'}) {
      print STDERR <<"DONE";
$functionality
requires the $packageName Perl package.
Trying to install this package (and any dependencies) automatically
using the ActivePerl Perl Package Manager (PPM) ...

  Executing: $cmd

Package installation could take a couple of minutes.

DONE
    }

    my $result = system($cmd);
    
    eval "require $packageName";
    if ($@) {
      print STDERR <<"DONE";

Attempt to install $packageName Perl package failed. Possible reasons
include: (1) Unable to install to Perl package area (you may need
Administrator permissions), (2) PPM is not on the System search path,
(3) Using Strawberry Perl instead of ActiveState Perl, or (4) there is
no network connection. The error code returned from executing
\"$cmd\" was $result.

Try installing the $packageName Perl package manually.
   
DONE
      exit($EXIT_NEED_PERL_PACKAGE);
    }
  }
}